eval 'exec perl -S -w $0 ${1+"$@"}'
        if 0;

################################################
#
# PLAYLIST
# (C) 1997-2000 by
#  Adam Lorentzon <d93-alo@nada.kth.se> and
#  LaLa <lala@c64.org>
#
# Requires:
#
#  - Perl5
#  - Un*x
#
# Usage:
#
#  playlist -o [normal|random|n|r] -r [yes|no|repeat|norepeat|1|0] <playlist.spl|playlist.m3u>
#  OR
#  playlist -h   (for help)
#
#  PERL script to play a list of SID files under Un*x from a playlist
#  that was generated by SIDPlay for Win or Winamp. See help for details.
#
################################################

# Modify $SIDPLAY if sidplay is not in your PATH.
#
# NOTE: Make sure that you put the *FULL* path here, without any
# environment variables, otherwise this script might not be able
# to start a new sidplay after a song finishes!

# $SIDPLAY = '/home/imre/bin/sidplay';

# Modify $STILVIEW if STILView is not in your PATH.

$STILVIEW = '/home/imre/bin/stilview';

# Modify $SIDPLAY_OPTIONS if you want to pass in different options.

# $SIDPLAY_OPTIONS = '-f22000 -s -pc';
$SIDPLAY_OPTIONS = '-f48000 -q -O1';

require 5;

# Required for flushing STDOUT.
use IO::Handle;
use Getopt::Std;

################################################
#
# SUBS
#
################################################

# Unixize - to get rid of annoying DOS chars.

sub unixize {
    $_[0] =~ s/&/\\&/g;
    $_[0] =~ s/'/\\'/g;
    $_[0] =~ s/\\/\//g;
}

# Function to pretty print.

sub printspec {
    my $text = $_[0];
    my $maxlength = 50;
    my $filler = '-';
    my $i;

    $i = int(($maxlength-length($text)-2)/2);

    print "\e[1m";
    print $filler x $i;
    print "\e[7m $text \e[0m\e[1m";

    $i = ($maxlength-$i-length($text)-2);

    print $filler x $i;
    print "\e[0m\n";
}

sub printhelp {
    print <<"HELP";
playlist.pl [-o normal|random] [-r yes|no] [-h] <playlist.spl|playlist.m3u>
Plays a SIDPlay/Win or Winamp generated playlist.
-h : this help
-o : [normal|n] plays the entries on the playlist one by one.
     [random|r] plays the entries in random order, not selecting those again
     that have already been played.
     This option overrides the playlist\'s setting.
-r : [no|n|0|norepeat] plays the playlist once, then quits.
     [yes|y|1|repeat|r] plays the playlist over and over.
     This option overrides the playlist\'s setting.

\$HVSC_BASE should be set to point to the HVSC base directory,
 or the BaseDir= line in the playlist should be set to it.

\$SIDPLAY should be set to the location of the sidplay executable,
 or sidplay should be in your \$PATH,
 or set \$SIDPLAY in this script.
 (SIDPLAY options are set in this script in \$SIDPLAY_OPTIONS!)

\$STILVIEW may be set to the location of the STILView executable,
 or STILView may be in your \$PATH,
 or set \$STILVIEW in this script,
 or don\'t set it at all - STILView is optional.

Examples:
playlist playlist.spl               - play the playlist with its own options
playlist playlist.spl -o random     - plays the playlist in random order
playlist playlist.m3u -r n          - plays the Winamp playlist only once,
                                      then exits (default behavior)
playlist playlist.spl -o r -r r     - plays the playlist in random order
                                      and continuously
HELP
    exit();
}

################################################
#
# MAIN PROGRAM STARTS HERE
#
################################################

getopts('ho:r:', \%opts);

if ((!defined(%opts) and !$ARGV[0]) or (defined($opts{h}))) {
    printhelp();
}

if (defined($opts{o})) {
    if (($opts{o} =~ /^\s*n\s*$/i) or ($opts{o} =~ /^\s*normal\s*$/i)) {
        $ORDER = "normal";
    }
    elsif (($opts{o} =~ /^\s*r\s*$/i) or ($opts{o} =~ /^\s*random\s*$/i)) {
        $ORDER = "random";
    }
    else {
        die "ERROR: Unknown -o option: $opts{o}!\n";
    }
}

if (defined($opts{r})) {
    if (($opts{r} =~ /^\s*[yr]\s*$/i) or ($opts{r} =~ /^\s*yes\s*$/i) or
        ($opts{r} =~ /^\s*1\s*$/i) or ($opts{r} =~ /^\s*repeat\s*$/i)) {
        $REPEAT="yes";
    }
    elsif (($opts{r} =~ /^\s*n\s*$/i) or ($opts{r} =~ /^\s*no\s*$/i) or
           ($opts{r} =~ /^\s*0\s*$/i) or ($opts{r} =~ /^\s*norepeat\s*$/i)) {
        $REPEAT="no";
    }
    else {
        die "ERROR: Unknown -r option: $opts{o}!\n";
    }
}

if (!$ARGV[0]) {
    die "ERROR: Playlist's filename must be defined!\n";
}


if ($ARGV[0] =~ /\.m3u\s*$/i) {
    $playlistformat = "m3u";
}
elsif ($ARGV[0] =~ /\.spl\s*$/i) {
    $playlistformat = "spl";
}
else {
    die "ERROR: Playlist must have .spl or .m3u extension!\n";
}

if (!open (PLAYLIST, $ARGV[0])) {
    die "ERROR: Couldn't open playlist '$ARGV[0]'!\n";
}

# Try to extract the HVSC_BASE env. var.

$basedir = $ENV{'HVSC_BASE'};
$basedir =~ s/[\/\\]\s*$//; # Removes trailing slash.

# Try to find where SIDPLAY is.

$sidplay = $ENV{'SIDPLAY'};

if (!$sidplay) {

    # No SIDPLAY env var, try to find sidplay.

    $sidplay = qx{ which sidplay };
    $sidplay =~ s/\n//g;
    $sidplay = $SIDPLAY if ($sidplay =~ /^no sidplay in/i);

    if (!$sidplay) {
        die "ERROR: Couldn't find SIDPLAY!\n";
    }
    $sidplay = join(' ',$sidplay,$SIDPLAY_OPTIONS);
}

# Try to find where STILVIEW is.

$stilview = $ENV{'STILVIEW'};

if (!$stilview) {

    # No STILVIEW env var, try to find STILView.

    $stilview = qx{ which stilview };
    $stilview =~ s/\n//g;
    $stilview = $STILVIEW if ($stilview =~ /^no stilview in/i);

    if (!$stilview) {
        # This is not an error - STILView is optional.
        print "WARNING: Couldn't find STILVIEW!\n";
    }
}

# Build up the playlist

$orderno=0;
$format=1; # Assume format 1.
while (<PLAYLIST>) {

    if ($playlistformat eq "m3u") {

        # Winamp playlist.

        if (/^\"(\S+)\"\s*(\d+)*\s*((\d+):(\d+))*/) {
            $file = $1;
            $subsong = $2;
            if ($4) {
                $time = $4*60+$5;
            }
            else {
                $time = 0;
            }
            $totalTime += $time;

            # Replace HVSC_BASEDIR with actual one...

            if (($file =~ /^HVSC_BASE/) and $basedir) {
                $file =~ s/^HVSC_BASE/$basedir/;
            }

            unixize($file);
            $playlist[$orderno] = "$file:$subsong:$time";
            $orderno++;
        }
    }

    else {

        # Sidplay playlist

        # Format

        if (/^Format=(\d+)/i) {
            $format = $1;
        }

        # BaseDir

        elsif (/^BaseDir=\s*(\S*)/i) {
            if (!$basedir) {
                $basedir = $1;
                unixize($basedir);
            }
        }

        # Order

        elsif (/^Order=\s*(\S*)/i) {
            $order = $1;
        }

        # Repeat

        elsif (/^Repeat=\s*(\S*)/i) {
            $repeat = $1;
        }

        if ($format == 1) {

            # Filename

            if (/^Filename=\s*(\S*)/) {
                $subsong = 0;
                $file = $1;
                unixize($file);
            }

            # Subsong #

            elsif (/^Subsong=\s*(\d*)/) {
                $subsong = $1;
            }

            # Time

            elsif (/^Time=\s*(\d*)/) {
                $time = $1;
            }

            # Order #

            elsif (/^\[(\d+)\]/) {
                $orderno = $1;
                $orderno--;
                if ($file) {
                    # We are putting the *previous* one on the list.
                    $playlist[$orderno-1] = "$file:$subsong:$time";
                    $totalTime += $time;
                }
                $file = "";
            }
        }
        else { # Must be format 2

            # Filename

            if (/^file=\s*(\S*)/) {
                $newfile = $1;
                unixize($newfile);

                # Now we insert the previous one to the playlist.
                if ($file) {
                    $playlist[$orderno] = "$file:$subsong:$time";
                    $totalTime += $time;
                    $orderno++;
                }
                $file = $newfile;
                $subsong = 0;
            }

            # Subsong #

            elsif (/^subtune=\s*(\d*)/) {
                $subsong = $1;
            }

            # Time

            elsif (/^time=\s*(\d*)/) {
                $time = $1;
            }
        }
    }
}

if ($playlistformat eq "spl") {
    # Insert the last one, too.
    if ($file) {
        $maxnosongs = $orderno+1;
        $playlist[$orderno] = "$file:$subsong:$time";
        $totalTime += $time;
    }
}
else {
    # Winamp playlist.
    $maxnosongs = $orderno;
}

# DEBUG
# $j=0;
# foreach $play (@playlist) {
#    print "$j - $play\n";
#    $j++;
# }
# print "$maxnosongs\n";

close PLAYLIST;

# Some defaults for Winamp playlists.

if (!defined($order)) {
    $order = "normal";
}

if (!defined($repeat)) {
    $repeat = "no";
}

# Order override.

if ($ORDER) {
    $order = $ORDER;
}

# Repeat override.

if ($REPEAT) {
    $repeat = $REPEAT;
}

if ($playlistformat eq "spl") {
    if ((!$basedir) or (!-e $basedir)) {
        die "ERROR: Couldn't find/determine HVSC base directory!\n";
    }
}

$totalhr = $totalTime/3600;
$totalmin = ($totalTime%3600)/60;
$totalsec = $totalTime%60;

# Build up the order of play.

if ($order =~ /normal/i) {
    for ($i = 0; $i < $maxnosongs; $i++) {
        $playorder[$i] = $i;
    }
}

# Play the playlist.

do {
    # If necessary, randomize playorder every time the playlist repeats.

    if ($order =~ /random/i) {

        local(@alltunes) = 0..$maxnosongs-1;
        @playorder = ();

        while (@alltunes) {
            push(@playorder, splice(@alltunes, rand(@alltunes), 1));
        }
    }

    $i = 0;
    do {
        ($file, $subsong, $time) = split (/:/, @playlist[$playorder[$i]]);

        if ($playlistformat eq "m3u") {
            $tempfile = $file;
            if ($file =~ /^$basedir/) {
                $file =~ s/^$basedir//;
            }
        }
        else {
            $tempfile = "$basedir$file";
        }

        if ($subsong) {
            $command = "$sidplay -o$subsong $tempfile";
        }
        else {
            $command = "$sidplay $tempfile";
        }

        # Playing time is not defined - enable input interface.
        if (!$time) {
            $input_interface = 1;
        }
        else {
            $input_interface = 0;
        }

        # DEBUG
        # print "$command\n";

        # Fancy output.

        print `clear`;

        print "\e[1m\e[7m------------------- PLAYLIST.PL ------------------\e[0m\n";
        print "\e[1m(C) 1997-2000 by Adam Lorentzon <d93-alo\@nada.kth.se>\n";
        print "              and LaLa <lala\@c64.org>\e[0m\n";
        printspec "PLAYLIST";
        print "Filename\t: $ARGV[0]\n";
        printf "Now playing\t: %d of %d\n", $i+1, $maxnosongs;
        print "HVSC base dir\t: $basedir\n";
        print "Playing order\t: $order\n";
        print "Total playtime\t: ";
        if ($totalhr != 0) {
            printf "%d hr ", $totalhr;
        }
        printf "%d:%02d\n", $totalmin, $totalsec;
        print "Play repeatedly\t: $repeat\n";

        printspec "CURRENT TUNE";
        printf "%s #%d\n", $file, $subsong;

        if ($time) {
            $min = $time/60;
            $sec = $time%60;
            printf "Length\t\t: %d:%02d\n", $min, $sec;
        }

        if ($stilview) {
            my $temp;
            printspec "STIL";
            $temp = `$stilview -e=$file -l=$basedir -b -o`;
            print "---- GLOBAL  COMMENT ----\n$temp" if $temp;
            $temp = `$stilview -e=$file -t=0 -f=comment -l=$basedir -b -s`;
            print "------ TUNE COMMENT -----\n$temp" if $temp;
            $temp = `$stilview -e=$file -t=$subsong -l=$basedir -s -b`;
            print "------ STIL  ENTRY ------\n$temp" if $temp;
            $temp = `$stilview -e=$file -t=$subsong -l=$basedir -s -o`;
            print "------- BUG ENTRY -------\n$temp" if $temp;
        }

        printspec "SIDPLAY";

        # Play the file until time is up.

        if ($pid = fork) {
            # Parent

            # Sleep is needed to let SIDPlay dump its output to STDOUT.
            # Otherwise, printouts get messed up.
            $time_left = $time-1;
            sleep(1);

            # Some tunes are only 1 sec long...
            if ($time_left != 0) {

                printspec "EXTRA";

                if ($input_interface) {
                    print "Hotkeys enabled.\n";
                    print "Play (N)ext tune, Play (S)ame tune, Play (P)revious tune\n";
                }
                else {
                    print "Time left: xx:xx";
                }

                do {
                    #
                    # Input thing...
                    #
                    # I'd love to have the input interface when the
                    # "Time left" thing is going, but for some odd
                    # reason, STDIN messes up the STDOUT flush, so... :-(

                    if ($input_interface) {

                        $input = <STDIN>;

                        # Skip to next tune.
                        if ($input =~ /n/i) {
                            $time_left = 1;
                        }

                        # Play same tune again.
                        elsif ($input =~ /s/i) {
                            $time_left = 1;
                            $i--;
                            if ($i == -1) {$i=$maxnosongs-2};
                        }

                        # Play previous tune.
                        elsif ($input =~ /p/i) {
                            $time_left = 1;
                            $i -= 2;
                            if ($i == -2) {$i=$maxnosongs-2};
                        }
                    }

                    #
                    # Time left output...
                    #

                    else {

                        $min_left = $time_left/60;
                        $sec_left = $time_left%60;
                        printf "\b\b\b\b\b%2d:%02d", $min_left, $sec_left;
                        STDOUT->flush();
                    }

                    sleep(1);
                    $time_left--;

                } until ($time_left == 0);
            }

            kill 9, $pid;
        }
        elsif (defined $pid) {
            # Child
            exec "$command";
        }
        else {
            # Weird fork error.
            printspec "ERROR!";
            die "Couldn't fork: $!\n";
        }
        $i++;
    } while ($i < $maxnosongs)
} until ($repeat =~ /no/i);

